import platform
import sys
from pathlib import Path
from platformdirs import PlatformDirs
import pandas as pd
import pendulum
import typer
from loguru import logger

from . import __version__


def setup_logger(level="INFO") -> Path:
    """ロガーの設定"""

    format_short = (" | ").join(
        ["{time:YYYY-MM-DDTHH:mm:ss}", "<level>{level:8}</level>", "<level>{message}</level>"]
    )
    format_long = (" | ").join(
        [
            "{time:YYYY-MM-DDTHH:mm:ss}",
            "<level>{level:8}</level>",
            "<cyan>{name}.{function}:{line}</cyan>",
            "<level>{message}</level>",
        ]
    )

    # ロガーをリセット
    logger.remove()

    # stderr用
    if level in ["DEBUG"]:
        logger.add(
            sys.stderr,
            format=format_long,
            level=level,
        )
    else:
        logger.add(
            sys.stderr,
            format=format_short,
            level=level,
        )

    # ファイル出力用
    p = PlatformDirs(appname="haniwers", version=__version__)
    fname = p.user_log_path / "haniwers_log.json"
    logger.add(
        sink=fname,
        format=format_long,
        level="DEBUG",
        serialize=True,
        retention="10 days",
        rotation="1 MB",
    )
    return fname


app = typer.Typer()


@app.command()
def version(env: bool = False, log_level: str = "INFO"):
    """Show version."""

    fname = setup_logger(level=log_level)

    msg = f"haniwers {__version__}"
    print(msg)
    logger.debug(msg)

    if env:
        msg = f"Python: {platform.python_version()}"
        print(msg)
        logger.debug(msg)

        msg = f"System: {platform.system()}"
        print(msg)
        logger.debug(msg)

        msg = f"OS: {platform.platform()}"
        print(msg)
        logger.debug(msg)

        msg = f"Logs: {fname}"
        print(msg)
        logger.debug(msg)

        print("\n🤖 Please use 'poetry env info' to see more details.\n")

    return


@app.command()
def raw2tmp(
    read_from: str,
    search_pattern: str = "*.csv",
    interval: int = 600,
    offset: int = 0,
    tz: str = "UTC+09:00",
    log_level: str = "INFO",
) -> None:
    """Parse raw_data into CSV format. Should be used temporarily for quick analysis.

    宇宙線を測定したデータをその場で確認するために、生データを変換するための簡易コマンドです。
    必要最低限のオプションしか変更できないようになっています。
    出力ファイル名（``tmp_raw2tmp.csv``）は固定で、変更できません。
    きちんとしたデータ解析には ``haniwers run2csv`` を使って変換してください。

    :Args:
    - read_from (str): 測定データがあるディレクトリ名
    - search_pattern (str, optional): 検索パターン. Defaults to "*.csv".
    - interval (int, optional): リサンプルの間隔. Defaults to 600.
    - offset (int, optional): 測定時刻のオフセット. Defaults to 0.
    - tz (str, optional): 測定時刻のタイムゾーン. Defaults to "UTC+09:00".
    """
    from .preprocess import get_fnames, raw2csv

    setup_logger(level=log_level)

    logger.info(f"Read data from {read_from}")
    fnames = get_fnames(read_from, search_pattern)
    gzip, csv = raw2csv(fnames, interval, offset, tz)
    logger.debug(f"raw2gz = {len(gzip)}")
    logger.debug(f"raw2csv = {len(csv)}")

    fname = "tmp_raw2tmp.csv.gz"
    gzip.to_csv(fname, index=False, compression="gzip")
    logger.info(f"Save data to: {fname} ({len(gzip)} rows).")

    fname = "tmp_raw2tmp.csv"
    csv.to_csv(fname, index=False)
    logger.info(f"Save data to: {fname} ({len(csv)} rows).")


@app.command()
def run2csv(
    runid: int, save: bool = False, load_from: str = "config.toml", log_level: str = "INFO"
) -> None:
    """Parse raw_data into CSV format. Specify RunID.

    :Args:
    - runid (int): ラン番号
    - save (bool, optional): 保存ファイル名（CSV形式）. Defaults to False.
    - load_from (str, optional): RunDataの設定ファイル. Defaults to "config.toml".
    """
    from .config import Config
    from .preprocess import run2csv

    setup_logger(level="INFO")
    c = Config(load_from)
    msg = f"Load config from: {load_from}."
    logger.info(msg)

    msg = f"Get RunData: {runid}."
    logger.info(msg)

    run = c.get_run(runid)
    logger.info(f"description: {run.description}")
    logger.info(f"read_from: {run.read_from}")
    logger.debug(f"srcf: {run.srcf}")

    gzip, csv = run2csv(run)
    if save:
        fname = run.raw2gz
        gzip.to_csv(fname, index=False, compression="gzip")
        logger.info(f"Save data to: {fname} ({len(gzip)} rows).")
        fname = run.raw2csv
        csv.to_csv(fname, index=False)
        logger.info(f"Save data to: {fname} ({len(csv)} rows).")
    else:
        logger.warning("No data saved. Add --save to save data.")
        logger.debug(f"gzip: {len(gzip)}.")
        logger.debug(f"csv:  {len(csv)}.")


@app.command()
def ports(log_level: str = "INFO") -> None:
    """Search available ports and show device names.

    :Note:
    - Linuxの場合: `/dev/ttyUSB0`
    - macOSの場合: `/dev/cu.usbserial-*` （CP2102N USB to UART Bridge Controller）
    - Windowsの場合: `COM3`
    """

    setup_logger(level=log_level)

    from serial.tools import list_ports

    ports = list_ports.comports()
    n = len(ports)

    if n == 0:
        logger.warning("No ports found")
        return

    logger.info(f"Found {n} ports")

    for i, port in enumerate(ports):
        logger.info(f"Port{i}: {port}")

        logger.debug(f"{port.device=}")
        logger.debug(f"{port.name=}")
        logger.debug(f"{port.description=}")
        logger.debug(f"{port.usb_description()=}")
        logger.debug(f"{port.hwid=}")
        logger.debug(f"{port.usb_info()=}")
        logger.debug(f"{port.pid=}")
        logger.debug(f"{port.vid=}")
        logger.debug(f"{port.interface=}")
        logger.debug(f"{port.manufacturer=}")
        logger.debug(f"{port.product=}")
        logger.debug(f"{port.serial_number=}")

    for port in ports:
        if "UART" in port.description:
            logger.info(f"Please use '{port.device}' as your device path")


@app.command()
def vth(
    ch: int = 0,
    vth: int = 0,
    max_retry: int = 3,
    load_from: str = "daq.toml",
    log_level: str = "INFO",
) -> None:
    """Set threshold to each channel. 1step = 4mV.

    OSECHIの各チャンネルにスレッショルドを設定します。
    カレントディレクトリにある ``thresholds_latest.csv`` を参考に、
    スレッショルド値の最良推定値を設定します。
    1step = 4mVに相当します。

    原因がまだわかってないですが、スレッショルドの書き込みに失敗することがあります。
    その場合は、書き込みに成功するまで``max_retry``回リトライします。デフォルトは3回です。

    :Args:
    - ch (int, optional): チャンネル番号. Defaults to 0 (=all).
    - vth (int, optional): スレッショルド値. Defaults to 0 (=all).
    - max_retry (int, optional): スレッショルドの書き込みに失敗したときにリトライする回数. Defaults to 3.
    - load_from (str, optional): 設定ファイル. Defaults to "daq.toml".
    """

    from .config import Daq
    from .daq import set_vth_retry

    setup_logger(level=log_level)

    daq = Daq()
    daq.load_toml(load_from)

    now = pendulum.now().format("YYYYMMDD")
    daq.saved = str(Path(daq.saved) / now)

    # 個別のチャンネルにスレッショルドを設定する
    if ch in range(1, 4) and vth > 0:
        logger.debug(f"Set threshold to each channel: {ch} -> {vth}")
        set_vth_retry(daq, ch, vth, max_retry)
        return

    # 引数を指定しない場合は
    # すべてのチャンネルに規定のスレッショルドを設定する
    if ch == 0 and vth == 0:
        # スレッショルド値をファイルから読み込む
        fname = Path("thresholds_latest.csv")
        if not fname.exists():
            logger.error(f"No file found. Please create {fname}")
            return

        names = ["ch", "3sigma"]
        thresholds = pd.read_csv(fname)[names]

        for _, row in thresholds.iterrows():
            ch = int(row["ch"])
            vth = int(row["3sigma"])
            logger.debug(f"Set threshold to channels: {ch} -> {vth}")
            set_vth_retry(daq, ch, vth, max_retry)
        return

    # オプション指定が間違っている
    logger.error("Invalid arguments")
    return


@app.command()
def scan(
    ch: int = 0,
    duration: int = 10,
    step: int = 1,
    vmin: int = 250,
    vmax: int = 311,
    quiet: bool = False,
    load_from="scan.toml",
    log_level: str = "INFO",
) -> None:
    """スレッショルド曲線を測定する

    :Args:
    - ch (int, optional): チャンネル番号. Defaults to 0.
    - duration (int, optional): 1点あたりの測定時間（秒）. Defaults to 10.
    - step (int, optional): 測定間隔. Defaults to 1.
    - vmin (int, optional): 測定範囲（最小値）. Defaults to 250.
    - vmax (int, optional): 測定範囲（最大値）. Defaults to 311.
    - quiet (bool, optional): _description_. Defaults to False.
    - load_from (str, optional): 設定ファイル. Defaults to "scan.toml".
    """
    from .config import Daq
    from .threshold import scan_thresholds

    setup_logger(level=log_level)
    daq = Daq()
    daq.load_toml(load_from)
    daq.quiet = quiet

    now = pendulum.now().format("YYYYMMDD")
    daq.saved = str(Path(daq.saved) / now)

    if ch == 0:
        channels = [1, 2, 3]
    else:
        channels = [ch]

    thresholds = list(range(vmin, vmax, step))

    for ch in channels:
        msg = f"Running threshold scan on ch{ch}."
        logger.info(msg)
        scan_thresholds(daq, duration, ch, thresholds)

    return


@app.command()
def daq(
    quiet: bool = False,
    load_from: str = "daq.toml",
    log_level: str = "INFO",
) -> None:
    """Start DAQ. Set up with daq.toml.

    :Args:
    - quiet (bool, optional): quiet mode. Defaults to False.
    - load_from (str, optional): 設定ファイル. Defaults to "daq.toml".
    """
    from .config import Daq
    from .daq import run, open_serial_connection

    # ログレベルを設定
    setup_logger(level=log_level)

    # DAQの初期設定
    args = Daq()
    args.load_toml(load_from)
    args.quiet = quiet

    # データの保存先をymdに変更
    now = pendulum.now().format("YYYYMMDD")
    args.saved = str(Path(args.saved) / now)

    with open_serial_connection(args) as port:
        run(port, args)

    return


@app.command()
def mock_daq(quiet: bool = False, load_from: str = "daq.toml", log_level: str = "DEBUG"):
    from unittest.mock import MagicMock, patch
    from .config import Daq
    from .mimic import FakeEvent
    from .daq import run
    import serial

    # ログレベルを設定
    setup_logger(level=log_level)
    logger.debug("mock-daq")

    # DAQの初期設定
    args = Daq()
    args.load_toml(load_from)
    args.quiet = quiet

    # データの保存先をymdに変更
    now = pendulum.now().format("YYYYMMDD")
    args.saved = str(Path(args.saved) / now)
    args.max_rows = 10
    args.max_files = 5
    logger.debug(args)

    # シリアル通信をモック
    mock_port = MagicMock()
    mock_port.readline().decode.return_value = FakeEvent().to_mock_string()
    mock_port.name.return_value = "mock"

    with patch("serial.Serial", return_value=mock_port):
        with serial.Serial() as port:
            logger.debug(f"Port opened: {port.name}")
            run(port, args)
    logger.debug(f"Port closed: {port.name}")


@app.command()
def fit(
    read_from: str,
    search_pattern: str = "threshold_scan.csv",
    ch: int = 0,
    log_level: str = "INFO",
):
    """Get threshold recommendations.

    スレッショルドを測定したデータから、誤差関数を使って閾値の推定値を計算します。
    チャンネル番号を指定できます。
    デフォルトはすべてのチャンネルの推定値を計算します。
    計算するたびにその結果は``thresholds_history.csv``に追記されます。
    また、最新のフィット結果は``thresholds_latest.csv``に保存されます。
    このファイルを使ってスレッショルドを設定できるようになっています。

    :Args:
    - read_from (str): スレッショルド測定データがあるディレクトリ名を指定してください
    - search_pattern (str, optional): スレッショルド測定データのファイル名を変更できます。Defaults to "threshold_scan.csv".
    - ch (int, optional): チャンネル番号を変更できます。Defaults to 0.
    """
    import pandas as pd

    from .preprocess import get_fnames
    from .threshold import fit_thresholds

    setup_logger(level=log_level)

    logger.info(f"Read data from {read_from}")
    fnames = get_fnames(read_from, search_pattern)

    # ファイルが見つからない時は、なにもしない
    if len(fnames) == 0:
        logger.error("No files found.")
        return

    logger.debug(fnames)

    # チャンネル番号が範囲外のときは、なにもしない
    if ch > 3:
        logger.error(f"Out of range!: {ch}")
        return

    channels = [ch]
    if ch == 0:
        channels = [1, 2, 3]

    names = ["time", "duration", "ch", "vth", "events", "tmp", "atm", "hmd"]
    data = pd.read_csv(fnames[0], names=names, parse_dates=["time"])
    thresholds = fit_thresholds(data, channels)

    # 実行した時刻を上書きする
    now = pendulum.now()
    thresholds["timestamp"] = now
    print(thresholds)

    fname = "thresholds_history.csv"
    thresholds.to_csv(fname, index=False, mode="a", header=None)
    logger.info(f"Saved to {fname}")
    fname = "thresholds_latest.csv"
    thresholds.to_csv(fname, index=False)
    logger.info(f"Saved to {fname}")

    return


if __name__ == "__main__":
    app()



import sys
from pathlib import Path

import pandas as pd
import polars as pl
from loguru import logger


from .config import RunData


def get_fnames(read_from: str, search_pattern: str) -> list[Path]:
    """測定データのファイル名の一番を取得

    測定データのファイルの一覧を取得します。
    ファイル名はPathオブジェクトに変換します。
    リストはファイル名で昇順ソートします。

    :Args:
    - read_from (str): 測定データがあるパス
    - search_pattern (str): ファイル名の検索パターン

    :Returns:
    - fnames (list[Path]): 測定データのファイル名（Pathオブジェクト）のリスト
    """
    fnames = sorted(Path(read_from).glob(search_pattern))
    return fnames


def read_data_with_pandas(fnames: list[Path]) -> pd.DataFrame:
    """pd.DataFrameに変換

    複数のファイルを読み込んでpd.DataFrameに変換します。

    :Args:
    - fnames (list[Path]): 測定データのファイル名（Pathオブジェクト）のリスト

    :Returns:
    - merged (pd.DataFrame): すべての測定データを結合したデータフレーム
    """
    names = ["datetime", "top", "mid", "btm", "adc", "tmp", "atm", "hmd"]
    data = []
    for fname in fnames:
        _suffix = fname.suffix
        if _suffix in [".dat"]:
            datum = pd.read_csv(fname, names=names, sep=" ", comment="t")
        elif _suffix in [".csv"]:
            datum = pd.read_csv(fname, names=names)
        else:
            msg = f"Unknown suffix: {_suffix}"
            logger.warning(msg)
            break
        data.append(datum)
    merged = pd.concat(data, ignore_index=True).dropna(how="all")

    if len(merged) == 0:
        msg = "No entries in DataFrame. Exit."
        logger.error(msg)
        sys.exit()

    return merged


def read_data_with_polars(fnames: list[Path]) -> pd.DataFrame:
    """pl.DataFrameに変換

    複数のファイルを読み込んでpl.DataFrameに変換します。
    polarsで読み込んだ方が、少し処理が速くなります。

    ただし、そのままpl.DataFrameを返すと、あとの処理がうまくいかないので、
    最後にpl.DataFrameからpd.DataFrameに変換しています。


    :Args:
    - fnames (list[Path]): 測定データのファイル名（Pathオブジェクト）のリスト

    :Returns:
    - merged (pd.DataFrame): すべての測定データを結合したデータフレーム
    """
    names = ["datetime", "top", "mid", "btm", "adc", "tmp", "atm", "hmd"]
    data = []
    for fname in fnames:
        _suffix = fname.suffix
        _stem = fname.stem
        if _suffix in [".dat"]:
            datum = pl.read_csv(
                fname,
                has_header=False,
                new_columns=names,
                separator=" ",
                comment_char="t",
            )
        elif _suffix in [".csv"]:
            if _stem.startswith("osechi_data"):
                datum = pl.read_csv(fname, has_header=False, new_columns=names)
            elif _stem.startswith("scan_data"):
                datum = pl.read_csv(fname, has_header=False, new_columns=names, skip_rows=1)
        else:
            msg = f"Unknown suffix: {_suffix}"
            logger.warning(msg)
            break
        data.append(datum)
    merged = pl.concat(data).to_pandas().dropna(how="all")

    if len(merged) == 0:
        msg = "No entries in DataFrame. Exit."
        logger.error(msg)
        sys.exit()

    return merged


def read_data(fnames: list[Path]) -> pd.DataFrame:
    """ファイルをデータフレームに変換

    ``read_data_with_polars``を使ってデータフレームに変換します。

    """
    # data = read_data_with_pandas(fnames)
    data = read_data_with_polars(fnames)
    return data


def add_time(data: pd.DataFrame, offset: int, timezone: str) -> pd.DataFrame:
    """カラムを追加

    測定データのデータフレームに``time``カラムを追加します。
    ``time``カラムは``datetime``カラムをタイムゾーン付き日時オブジェクトに変換したものです。
    また、``offset``（秒）の分、測定時刻を正しい時刻に修正します。

    :Args:
    - data (pd.DataFrame): 測定データ
    - offset (int): 測定時刻のズレ（オフセット）[秒]
    :Returns:
    - data (pd.DataFrame): ``time``カラムを追加した測定データ

    ```{caution}
    Raspberry Pi 4はRTCを内蔵していません。電源を切ると時計がストップします。
    測定を開始する前に、ネットワークに接続してNTPから時刻を自動取得するか、
    ``date``コマンドなどで時刻を手動設定する必要があります。
    その作業を忘れてしまった場合のために、``offset``オプションがあります。
    ```

    """
    data["datetime"] = pd.to_datetime(data["datetime"], format="ISO8601")
    if data["datetime"].dt.tz is None:
        data["datetime"] = data["datetime"].dt.tz_localize(timezone)
    data["datetime_offset"] = pd.to_timedelta(offset, "s")
    data["time"] = data["datetime"] + data["datetime_offset"]
    return data


def add_hit(data: pd.DataFrame) -> pd.DataFrame:
    """カラムを追加

    各シンチのレイヤーにヒットがあったかどうかを確認し、``hit_レイヤー名``カラムを追加します。
    ヒットがある場合は``1``、ヒットがない場合は``0``です。

    :Args:
    - data (pd.DataFrame): 測定データ

    :Returns:
    - data (pd.DataFrame): レイヤーごとのヒット情報を追加した測定データ
    pd.DataFrame
        ヒット情報を追加したデータです。
    """
    copied = data.copy()
    headers = ["top", "mid", "btm"]
    for header in headers:
        name = f"hit_{header}"
        copied[name] = 0
        isT = data[header] > 0
        copied.loc[isT, name] = 1
    return copied


def add_hit_type(data: pd.DataFrame) -> pd.DataFrame:
    """カラムを追加

    レイヤーごとのヒット情報から、ヒットの種類を計算します。
    ヒット情報が計算できてない場合は、警告を表示し、元のデータを返します。

    :Args:
    - data (pd.DataFrame): ``add_hit``したあとの測定データ

    :Returns:
    - data (pd.DataFrame): ``hit_type``を追加したデータフレーム
    """
    _header = "hit_top"
    if _header not in data.columns:
        msg = f"Header '{_header}' does not exist in current dataframe."
        logger.warning(msg)
        return data
    data["hit_type"] = 4 * data["hit_top"] + 2 * data["hit_mid"] + data["hit_btm"]
    return data


def resample_data(data: pd.DataFrame, interval: int):
    """時間を指定して測定データを再集計する

    ``data``に指定した測定データを``pandas.DataFrame.resample``で再集計します。
    データは前処理済みのデータを指定してください。
    再集計したデータをさらに、再集計することもできます。

    - 入力データに必要なカラム: ``["time", "hit_type", "hit_top", "hit_mid", "hit_btm", "adc", "tmp", "atm", "hmd"]
    - 出力データに追加されるカラム: ``["interval", "event_rate", "event_rate_top", "event_rate_mid", "event_rate_btm"]

    :Args:
        - data (pd.DataFrame): 読み込んだデータを指定。
        - interval (int): 再集計する時間間隔（秒）を指定

    :Returns:
        - merged (pd.DataFrame): 指定した時間間隔で再集計したデータ
    """

    # 再集計する時間間隔
    rule = f"{interval}s"

    # 元データのコピーを作成
    copied = data.copy()
    # インデックスを datetime オブジェクトに変更
    copied.index = copied["time"]

    # count
    headers = {"hit_type": "events"}
    keys = list(headers.keys())
    _count = copied.resample(rule)[keys].count().reset_index().rename(columns=headers)

    # sum: 合計値
    headers = {
        "hit_top": "hit_top_sum",
        "hit_mid": "hit_mid_sum",
        "hit_btm": "hit_btm_sum",
    }
    keys = list(headers.keys())
    _sum = copied.resample(rule)[keys].sum().reset_index()

    # mean: 平均値
    headers = {
        "adc": "adc_mean",
        "tmp": "tmp_mean",
        "atm": "atm_mean",
        "hmd": "hmd_mean",
    }
    keys = list(headers.keys())
    _mean = copied.resample(rule)[keys].mean().reset_index()

    # std : 標準偏差
    headers = {
        "adc": "adc_std",
        "tmp": "tmp_std",
        "atm": "atm_std",
        "hmd": "hmd_std",
    }
    keys = list(headers.keys())
    _std = copied.resample(rule)[keys].std().reset_index().rename(columns=headers)

    # 測定日時（time）で結合
    m1 = pd.merge(_count, _sum, on="time")
    m2 = pd.merge(m1, _mean, on="time")
    merged = pd.merge(m2, _std, on="time")

    # レートを計算
    merged["interval"] = interval
    merged["event_rate"] = merged["events"] / merged["interval"]
    merged["event_rate_top"] = merged["hit_top"] / merged["interval"]
    merged["event_rate_mid"] = merged["hit_mid"] / merged["interval"]
    merged["event_rate_btm"] = merged["hit_btm"] / merged["interval"]

    # 測定時間を計算
    epoch = merged["time"].min()
    merged["days"] = merged["time"] - epoch
    merged["seconds"] = merged["days"].dt.total_seconds()

    return merged


def resample_data_with_hit_type(data: pd.DataFrame, interval: int) -> pd.DataFrame:
    """
    ヒット種類ごとに時間で再集計します。

    :Args:
    - data (pd.DataFrame): ``hit_type``が追加されている測定データ
    - interval (int):  リサンプルの間隔 [秒]

    :Returns:
    - resampled (pd.DataFrame): ``hit_type``ごとにリサンプルしたデータ
    """
    hit_types = sorted(data["hit_type"].unique())
    rdata = []
    for ht in hit_types:
        q = f"hit_type == {ht}"
        qdata = data.query(q).copy()
        _resampled = resample_data(qdata, interval)
        _resampled["hit_type"] = ht
        rdata.append(_resampled)

    resampled = pd.concat(rdata, ignore_index=True)
    resampled = resampled.sort_values(["time", "hit_type"])
    return resampled


def preprocess_data(
    data: pd.DataFrame, interval: int, datetime_offset: int, timezone: str
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """測定データを前処理する

    :Args:
    - data (pd.DataFrame): 測定データ
    - interval (int): リサンプルする間隔 [秒]
    - datetime_offset (int): 時刻の修正 [秒]
    - timezone (str): タイムゾーン情報

    :Returns:
    - data (pd.DataFrame): 前処理したあとのデータ
    - resampled (pd.DataFrame): リサンプルしたデータ
    """
    data = add_time(data, offset=datetime_offset, timezone=timezone)
    data = add_hit(data)
    data = add_hit_type(data)
    resampled = resample_data_with_hit_type(data, interval)
    return data, resampled


def raw2csv(
    fnames: list[Path], interval: int, datetime_offset: int, timezone: str
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """測定データを指定してデータフレームに変換する

    :Args:
    - fnames (list[Path]): 前処理したいファイル名（Pathオブジェクト）のリスト
    - interval (int): リサンプルする間隔 [秒]
    - datetime_offset (int): 時刻の修正 [秒]
    - timezone (str): タイムゾーン情報

    :Returns:
    - data (pd.DataFrame): 前処理したあとのデータ
    - resampled (pd.DataFrame): リサンプルしたデータ

    """
    if len(fnames) == 0:
        msg = "No files found."
        logger.error(msg)
        sys.exit()
    data = read_data(fnames)
    data, resampled = preprocess_data(data, interval, datetime_offset, timezone)
    return data, resampled


def run2csv(run: RunData) -> tuple[pd.DataFrame, pd.DataFrame]:
    """ランを指定してデータフレームに変換する

    :Args:
    - run (RunData): RunDataオブジェクト

    :Returns:
    - data (pd.DataFrame): 前処理したあとのデータ
    - resampled (pd.DataFrame): リサンプルしたデータ

    """
    fnames = get_fnames(read_from=run.read_from, search_pattern=run.srcf)
    interval = run.interval
    datetime_offset = run.datetime_offset
    timezone = "UTC+09:00"

    data, resampled = raw2csv(fnames, interval, datetime_offset, timezone)
    data["runid"] = run.run_id
    data["name"] = run.name
    data["description"] = run.description

    resampled["runid"] = run.run_id
    resampled["name"] = run.name
    resampled["description"] = run.description

    return data, resampled


if __name__ == "__main__":
    fname = "../data/raw_data/20230806_run17/osechi_data_000000.csv"
    fnames = [Path(fname)]
    gzip, csv = raw2csv(fnames, interval=600, datetime_offset=0, timezone="UTC+09:00")
    logger.debug(f"raw2gz  = {len(gzip)}")
    logger.debug(f"raw2csv = {len(csv)}")



from dataclasses import dataclass
from icecream import ic
from loguru import logger
from pathlib import Path
from typing import Optional, Any
import pandas as pd
import sys
import tomli  # Python3.9対応のためtomliを使う

# import tomllib  # TODO: いずれtomllibに置き換える
from pydantic import BaseModel

ic.configureOutput(prefix=f"{__name__}: ")


@dataclass
class RunData:
    """
    ランの設定をまとめる用のクラス

    :Parameters:
    - run_id(int): ラン番号
    - read_from(str): 入力ディレクトリ名
    - srcf(str): 入力ファイル（の形式）
    - interval(int): サンプリング間隔（秒）。デフォルトは ``600`` （10分）
    - datetime_offset(int): 測定機器の時刻と実時刻の時間差（秒）。デフォルトは0秒
    - description(str): ランの簡単な説明。詳しい説明は別に用意したGoogleスプレッドシートに書き込む
    - skip(bool): 除外する場合は ``True`` にする。デフォルトは ``False``
    - nfiles(int): 読み込んだファイルの数。デフォルトは ``0``
    - raw2gz(str): リサンプルせず ``.csv.gz`` として書き出すァイル名。ファイル名を指定しない場合（ ``""`` or ``None`` ）は、ファイルに書き出さない
    - raw2csv(str): リサンプルして ``.csv`` として書き出すファイル名。ファイル名を指定しない場合（ ``""`` or ``None`` ）は、ファイルに書き出さない

    :Returns:
    - run(RunData): ランデータ

    """

    run_id: int
    description: str = ""
    read_from: str = ""
    srcf: str = "*.csv"
    interval: int = 600
    datetime_offset: int = 0
    skip: bool = False
    nfiles: int = 0
    raw2gz: str = ""
    raw2csv: str = ""
    query: str = ""

    def __post_init__(self) -> None:
        self.name = f"Run{self.run_id}"
        self.fnames = self.get_fnames()
        self.nfiles = len(self.fnames)

    def get_fnames(self) -> list[Path]:
        """該当するランのファイル名（ ``pathlib.Path`` ）の一覧を取得する

        :Returns:
        - fnames(list[Path]): ファイル名（Pathオブジェクト）のリスト

        :Notes:
        - 設定ファイルの ``srcd`` / ``srcf`` から該当するファイルの一覧を取得する
        - ``srcd`` が存在しない場合は終了する
        """

        read_from = Path(self.read_from).expanduser()
        srcf = self.srcf

        if not read_from.exists():
            error = f"No directory found : {read_from}"
            logger.error(error)
            sys.exit()

        fnames = sorted(read_from.glob(srcf))
        return fnames

    def _load_gzip(self) -> Optional[pd.DataFrame]:
        """gzipで保存したデータを読み込む

        ``raw2gz``で指定したファイル名をデータフレームとして読み込む。
        ファイルが存在しない場合は終了する。

        :Returns:
        - data(pd.DataFrame | None): データフレーム
        """
        p = Path(self.raw2gz)
        if p.exists():
            data = pd.read_csv(self.raw2gz, parse_dates=["time"])
            return data
        else:
            error = f"File not found. {p}"
            logger.error(error)
            sys.exit()

    def _load_csv(self) -> Optional[pd.DataFrame]:
        """csvで保存したデータを読み込む

        ``raw2csv``で指定したファイル名をデータフレームとして読み込む。
        ファイルが存在しない場合は終了する。

        :Returns:
        - data(pd.DataFrame | None) : データフレーム
        """
        p = Path(self.raw2csv)
        if p.exists():
            data = pd.read_csv(self.raw2csv, parse_dates=["time"])
            return data
        else:
            error = f"File not found. {p}"
            logger.error(error)
            sys.exit()

    def load_data(self, kind: str) -> Optional[pd.DataFrame]:
        """保存形式（"csv" or "gzip"）を指定してデータを読み込む

        :Parameters:
        - kind(str): 保存したデータの形式（"csv" or "gzip")

        :Returns:
        - data(pd.DataFrame): データフレーム
        """
        if kind == "csv":
            return self._load_csv()
        elif kind == "gzip":
            return self._load_gzip()
        else:
            fmt = ["csv", "gzip"]
            error = f"Wrong file type : {kind}. Choose from {fmt}."
            logger.error(error)
            sys.exit()


@dataclass
class Config:
    """設定ファイル用のクラス

    :Returns:
    - config(Config): 設定
    """

    fname: str = "config.toml"

    def __post_init__(self) -> None:
        """
        - ``self.config``
        - ``self.rules``
        - ``self.runs``
        """
        self.config = self.load_config()
        self.rules = self.get_rules()
        self.runs = self.get_runs()
        self.labels = self.get_labels()

    def load_config(self) -> dict:
        """設定ファイルを読み込む

        :Returns:
        - config(dict) : 設定

        :Notes:
        - デフォルトの設定ファイル名は `config.toml`
        - 設定ファイルの名前は変更することができる
        - 設定ファイルが見つからない場合は、エラーを表示してスキップ（早期リターン）する
        - 設定ファイルは辞書型で読み込む
        """
        p = Path(self.fname)

        if not p.exists():
            error = f"No file found : {p}"
            logger.error(error)
            sys.exit()

        with p.open("rb") as f:
            config = tomli.load(f)
        return config

    def get_rules(self) -> dict:
        """イベント条件に関する設定を取得する

        :Returns:
        - rules(dict): {イベント名 : イベント条件}

        :Notes:
        - 設定ファイルの ``[rules]`` セクションの内容を取得する
        - ``条件名 = 条件式`` の辞書型（map型）で定義されている

        """
        rules = self.config.get("rules")
        return rules  # type: ignore

    def get_labels(self) -> Optional[dict]:
        """カラム名を取得する

        ``[labels]`` のセクションに、カラム名に対応した日本語を記述する

        """
        labels = self.config.get("labels")
        return labels

    def get_run(self, run_id: int) -> RunData:
        """ラン番号を指定してラン情報を取得する"""
        for run in self.runs:
            if run.run_id == run_id:
                return run
        msg = f"Run #{run_id} is out of range. Add to config."
        logger.error(msg)
        sys.exit()

    def get_runs(self) -> list[RunData]:
        """ランに関係する設定を取得する

        :Returns:
        - runs(list[RunData]): ランごとに設定をまとめたリスト

        :Notes:
        - 設定ファイルの ``[[rawdata]]`` セクションの内容を ``RunData`` クラスにまとめる
        - 除外するランがある場合は ``skip = true`` を設定する
        """
        runs = []
        rundata = self.config["rundata"]
        for data in rundata:
            if data.get("skip") is None:
                run = data.get("run")
                error = f"Run{run} : No configuration found for 'skip'"
                logger.error(error)
                sys.exit()

            if not data.get("skip"):
                _data = RunData(
                    run_id=data["run"],
                    read_from=data["read_from"],
                    srcf=data["srcf"],
                    interval=data["interval"],
                    datetime_offset=data.get("datetime_offset", 0),
                    description=data["desc"],
                    skip=data["skip"],
                    raw2gz=data.get("raw2gz"),
                    raw2csv=data.get("raw2csv"),
                    query=data.get("query"),
                )
                runs.append(_data)
        return runs


@dataclass
class Daq:
    "Configuration of DAQ"

    saved: str = "."
    prefix: str = "data"
    suffix: str = ".csv"
    skip: int = 10
    max_rows: int = 10000
    max_files: int = 100
    quiet: bool = False
    append: bool = False
    device: str = "/dev/ttyUSB0"
    baudrate: int = 115200
    timeout: int = 1000
    fname_logs: str = "threshold_logs.csv"
    fname_scan: str = "threshold_scan.csv"

    def load_toml(self, fname: str) -> None:
        "Load DAQ configuration from TOML"
        p = Path(fname)
        with p.open("rb") as f:
            _config = tomli.load(f)

        # logger.debug(_config)
        self.saved = _config.get("saved", ".")
        self.prefix = _config.get("prefix", "data")
        self.suffix = _config.get("suffix", ".csv")
        self.skip = _config.get("skip", 10)
        self.max_rows = _config.get("max_rows", 10000)
        self.max_files = _config.get("max_files", 100)
        self.quiet = _config.get("quiet", False)
        self.append = _config.get("append", False)
        self.device = _config.get("device", "/dev/ttyUSB0")
        self.baudrate = _config.get("baudrate", 115200)
        self.timeout = _config.get("timeout", 1000)
        self.fname_logs = _config.get("fname_logs", "threshold_logs.csv")
        self.fname_scan = _config.get("fname_scan", "threshold_scan.csv")


class UserSettings(BaseModel):
    """ユーザー設定用のクラス

    ``read_from`` に指定したファイルからユーザー設定を読み込みます。
    設定ファイルはTOML形式で作成してください。
    その他の形式に対応する予定はいまのところありません。
    また、読み込み時のファイル形式のチェックもしていません。

    :versionadded: `0.12.0`

    :Example:

    ```python
    >>> s = UserSettings("../sandbox/hnw.toml")
    >>> s.settings
    {
        'default': {'saved': '', 'suffix': '.csv', 'skip': 10, 'max_rows': 1000},
        'device': {'port': '/dev/ttyUSB0', 'baudrate': 115200, 'timeout': 100},
        'daq': {'prefix': 'osechi_data', 'max_files': 1000},
        'scan': {'prefix': 'scan_data', 'max_files': 10, 'timeout': 10},
        'threshold': {'logs': {'fname': 'threshold_logs.csv',
        'names': ['time', 'ch', 'vth', 'success']},
        'scan': {'fname': 'threshold_scan.csv',
        'names': ['time', 'duration', 'ch', 'vth', 'events']
    }

    >>> s.sections
    dict_keys(
        ['default', 'device', 'daq', 'scan', 'threshold', 'loguru']
    )
    ```

    """

    load_from: str
    settings: dict = {}
    sections: list = []

    def model_post_init(self, __context: Any) -> None:
        settings = self.load_toml(self.load_from)
        self.settings = settings
        self.sections = list(settings.keys())
        return super().model_post_init(__context)

    def load_toml(self, load_from: str) -> dict:
        """設定をTOMLファイルから読み込んで、辞書型に変換する

        UserSettingsクラスのオブジェクトを生成するときに ``__post_init__``の中で実行している。
        プロダクション環境では、このメソッドをわざわざ実行する必要はありません。

        新しい設定ファイルを作成した場合の内容確認のために使えると思います。

        :Args:
        - load_from(str): ファイル名

        :Returns:
        - settings(dict): ユーザー設定

        ```

        """
        p = Path(load_from)
        with p.open("rb") as f:
            settings = tomli.load(f)
        return settings

    def _get_settings(self, keys: list) -> dict:
        """設定を取得する

        :Example:
        >>> us = UserSettings(load_from="../sandbox/hnw.toml")
        >>> keys = ["default", "device", "scan", "threshold"]
        >>> settings = s._get_settings(keys)
        """
        settings = {}
        for key in keys:
            d = self.settings.get(key)
            if d is None:
                pass
            else:
                settings.update(d)
        return settings

    def get_daq_settings(self) -> dict:
        """DAQに必要な設定を取得する

        :Returns:
        - settings(dict): DAQの設定に必要な項目

        :Example:

        ```python
        >>> us = UserSettings(load_from="../sandbox/hnw.toml")
        >>> settings = us.get_daq_settings()
        >>> settings
        {
            'saved': '',
            'suffix': '.csv',
            'skip': 10,
            'max_rows': 1000,
            'port': '/dev/ttyUSB0',
            'baudrate': 115200,
            'timeout': 100,
            'prefix': 'osechi_data',
            'max_files': 1000
        }
        ```

        """
        keys = ["default", "device", "daq"]
        settings = self._get_settings(keys)
        return settings

    def get_scan_settings(self) -> dict:
        """スレッショルド測定に必要は設定を取得する

        :Returns:
        - settings(dict): スレッショルド測定に必要な項目

        :Example:

        ```python
        >>> us = UserSettings(load_from="../sandbox/hnw.toml")
        >>> settings = us.get_scan_settings()
        >>> settings
        {
            'saved': '',
            'suffix': '.csv',
            'skip': 10,
            'max_rows': 1000,
            'port': '/dev/ttyUSB0',
            'baudrate': 115200,
            'timeout': 10,
            'prefix': 'scan_data',
            'max_files': 10,
            'logs': {'fname': 'threshold_logs.csv', 'names': ['time', 'ch', 'vth', 'success']},
            'scan': {'fname': 'threshold_scan.csv', 'names': ['time', 'duration', 'ch', 'vth', 'events']},
            'fit': {'fname': 'threshold_fit.csv', 'names': ['time', 'ch', '0sigma', '1sigma', '3sigma', '5sigma']}
        }
        ```

        """

        keys = ["default", "device", "scan", "threshold"]
        settings = self._get_settings(keys)
        return settings

    def get_loguru(self, level: str = "DEBUG") -> logger:
        """ロガーを初期化する

        loguruパッケージを使ったロガーを初期化します。
        ``level`` オプションで指定したログレベルで、標準エラー出力（``sys.stderr``）のハンドラーを作成します。

        以下のような ``[loguru]`` セクションを作成することで、ファイル出力のハンドラーを追加できます。
        このハンドラーのログレベルは``DEBUG``、出力形式はJSON形式にハードコードしています。

        ```toml
        [loguru]
        sink = "required"
        format = "required"
        retention = "optional"
        rotation = "optional"
        ```

        :Args:
        - level(str): 標準エラー出力のログレベル

        :Returns:
        - logger(loguru.logger): ロガー

        :Example:

        ```python
        >>> s = UserSettings("../sandbox/hnw.toml")
        >>> logger = s.get_loguru()
        >>> logger.info("ロガーを初期化した")
        ```
        """
        # ロガーの既定値をリセット
        logger.debug("ロガーをリセットする")
        logger.remove()

        # 標準エラー（sys.stderr）出力のハンドラーを追加
        logger.add(
            sys.stderr,
            format="{time:YYYY-MM-DDTHH:mm:ss} | <level>{level:8}</level> | <level>{message}</level>",
            level=level,
        )
        logger.debug(f"標準エラー出力のハンドラーを追加した（{level=}）")

        # ファイル出力のハンドラーを追加
        section = self.settings.get("loguru")

        if section is None:
            return logger

        # ファイルから読み込んだ値を設定
        logger.add(
            sink=section.get("sink"),
            format=section.get("format"),
            level="DEBUG",
            retention=section.get("retention"),
            rotation=section.get("rotation"),
            serialize=True,
        )
        logger.info("ファイル出力のハンドラーを追加した")
        return logger


if __name__ == "__main__":
    """configモジュールの動作確認用

    $ python3 config.py

    - 設定ファイルの内容がきちんと読み込まれているか確認する
    - Configクラスのインスタンス変数を修正した場合に動作確認する
    - RunDataクラスのインスタンス変数を修正した場合に動作確認する
    """

    c = Config("../sandbox/config.toml")
    ic(c.fname)
    ic(type(c.rules))
    # ic(c.runs)
    # for run in c.runs:
    #     ic(run.name)
    #     ic(run.runid)
    #     ic(run.fnames)



import csv
import sys
from datetime import datetime
from pathlib import Path

import pandas as pd
import pendulum
import serial
from deprecated import deprecated
from loguru import logger
from pydantic import BaseModel
from tqdm import tqdm
from typing import TextIO, Generator

from .config import Daq


class RealEvent(BaseModel):
    """実イベント

    OSECHIに接続したUSBポートから、シリアル通信で受け取った値を格納するためのデータクラス。
    ファイルに書き出したり、`pandas.DataFrame`に変換できるように自作メソッドを追加。

    """

    timestamp: datetime = pendulum.now()
    """測定時刻。宇宙線イベントが通過した日時。タイムゾーン付きの日付オブジェクト"""
    top: int = 0
    """topレイヤーのヒット。0 - 10の値"""
    mid: int = 0
    """midレイヤーのヒット。0 - 10の値"""
    btm: int = 0
    """btmレイヤーのヒット。0 - 10の値"""
    adc: int = 0
    """topレイヤーにヒットがあったときのADC値。0 - 1023の値"""
    tmp: float = 0.0
    """BME280で測定した気温。[degC]"""
    atm: float = 0.0
    """BME280で測定した気圧。[Pa]"""
    hmd: float = 0.0
    """BME280で測定した湿度。[%]"""

    def to_list_string(self) -> list[str]:
        """List

        メンバー変数を文字列にしたリストに変換します。

        ```python
        >>> real_data = RealEvent()
        >>> read_data.to_list_string()
        ['2024-05-21 08:44:20.389786+09:00', '0', '0', '0', '0', '0', '0', '0']
        ```

        """
        data = self.model_dump()
        values = [str(v) for v in data.values()]
        return values

    def to_csv_string(self) -> str:
        """Comma Separated Values

        メンバー変数をCSV形式（カンマ区切り）の文字列に変換します。
        OSECHIから受け取ったデータを、ファイルに保存する際に使うことを想定したメソッドです。

        ```python
        >>> real_data = RealEvent()
        >>> real_data.to_csv_string()
        '2024-05-21 08:44:20.389786+09:00,0,0,0,0,0,0,0'
        ```

        """
        data = self.model_dump().values()
        values = [str(v) for v in data]
        csv_string = (",").join(values)
        return csv_string

    def to_tsv_string(self) -> str:
        """Tab Separated Values

        メンバー変数をTSV形式（タブ区切り）の文字列に変換します。

        ```python
        >>> real_data = RealEvent()
        >>> real_data.to_tsv_string()
        '2024-05-21 08:44:20.389786+09:00\t0\t0\t0\t0\t0\t0\t0'
        ```

        """
        data = self.model_dump().values()
        values = [str(v) for v in data]
        tsv_string = ("\t").join(values)
        return tsv_string

    def to_ltsv_string(self) -> str:
        """Labeled Tab-Separated Values

        メンバー変数をLTSV形式（ラベルありのタブ区切り）の文字列に変換します。

        ```python
        >>> real_data = RealEvent()
        >>> real_data.to_ltsv_string()
        'timestamp:2024-05-21 08:44:20.389786+09:00\ttop:0\tmid:0\tbtm:0\tadc:0\ttmp:0\tatm:0\thmd:0'
        ```
        """
        data = self.model_dump().items()
        values = [f"{k}:{v}" for k, v in data]
        ltsv_string = ("\t").join(values)
        return ltsv_string


@deprecated(version="0.15.0", reason="Use mkdir_saved")
def init_saved(daq: Daq) -> None:
    """（削除予定）"""
    logger.warning("Deprecated since 0.15.0: Use mkdir_saved")
    return mkdir_saved(daq)


def mkdir_saved(daq: Daq) -> None:
    """データを保存するディレクトリを作成

    データを保存するディレクトリを初期化します。
    ディレクトリが存在しない場合は、新しく作成します。
    ディレクトリが存在する場合は、そのままにします。

    :Args:
    - daq(Daq): DAQ設定オブジェクト
    ```
    """

    p = Path(daq.saved)
    p.mkdir(exist_ok=True)
    msg = f"Save files to : {p}"
    logger.info(msg)

    return


def get_savef(args: Daq, fid: int | str) -> Path:
    """（削除予定）データを保存するファイル名を生成する

    保存するファイル名を生成します。
    DAQ設定の接頭辞（``prefix``）と拡張子（``suffix``）の値を使って
    ``{prefix}_{連番:06}.{suffix}``の形式で生成します。

    :Args:
    - args(Daq): DAQ設定オブジェクト
        - prefix: ファイルの接頭辞
        - suffix: ファイルの拡張子
    - n(int): ファイル番号

    :Example:
    ```console
    osechi_data_000000.csv
    osechi_data_000001.csv
    osechi_data_000002.csv
    ```

    """
    stem = f"{args.prefix}_{fid:07}"
    fname = Path(stem).with_suffix(args.suffix)
    savef = Path(args.saved, fname)

    msg = "deprecation warning: use get_savef_with_timestamp instaed."
    logger.warning(msg)

    return savef


def get_savef_with_timestamp(args: Daq, fid: int | str) -> Path:
    """データを保存するファイル名を生成する（時刻付き）

    作成日を含んだファイル名を生成する。
    ファイル名は、DAQ設定の接頭辞（``prefix``）、ファイルを開いた時刻（``pendulum.now``）と
    拡張子（``suffix``）の値を使って生成する。

    時刻のフォーマットは、ファイル名が分かりやすいように独自フォーマットにした。

    :Args:
    - `args (Daq)`: Daqオブジェクト
    - `fid (int|str)`: ファイル識別子

    :Returns:
    - `savef (Path)`: ファイル名（Pathオブジェクト）

    :Examples:
    ```console
    20240520/osechi_data_2024-05-20T12h34m56s_000000.csv
    20240520/osechi_data_2024-05-20T13h53m24s_000001.csv
    20240520/osechi_data_2024-05-20T14h46m23s_000000.csv  // 走り直すとリセット
    20240520/osechi_data_2024-05-20T14h36m32s_000001.csv
    ```

    """
    ts = pendulum.now().format("YYYY-MM-DDTHH[h]mm[m]ss[s]")
    # fidはintもしくはstrなので 07 とした
    # 07d だとstrのときにエラーがでる
    stem = f"{args.prefix}_{ts}_{fid:07}"
    fname = Path(stem).with_suffix(args.suffix)
    savef = Path(args.saved, fname)
    return savef


def open_serial_connection(daq: Daq) -> serial.Serial:
    """シリアル通信を開始する

    シリアル通信（UART）に使うポートを準備します。
    ``with``構文で使う想定です。

    通信に使うUSBポート名（``device``）、
    ボーレート（``baudrate``）、
    通信開始／書き込みのタイムアウト秒（``timeout``）は
    DAQ用の設定ファイルで変更できるようにしてあります。

    :Args:
    - daq(Daq): DAQ設定オブジェクト
        - device: USBポート名
        - baudrate: ボーレート（通信速度）[bps]
        - timeout: タイムアウト秒 [sec]

    :Returns:
    - port(serial.Serial): 通信を開始したシリアルオブジェクト

    :Example:

    ```python
    >>> with open_serial_connection(daq) as port:
    >>>     # データ測定の処理
    >>>
    ```

    """
    port = serial.Serial(
        daq.device,
        baudrate=daq.baudrate,
        timeout=daq.timeout,
        write_timeout=daq.timeout,
    )
    port.rts = False
    port.dtr = False

    # logger.debug(f"Port opened: {port}")

    return port


def write_vth(port: serial.Serial, ch: int, vth: int):
    """Write threshold to OSECHI.

    シリアル通信（UART）を使って閾値を書き込みます。
    値を書き込んだあとに、値が読み出せるかを確認します。

    読み出した値が ``dame`` となっている場合は書き込みに失敗しています。
    その場合は、もういちど設定し直すよう警告を表示します。

    ESP32のバッファサイズの制限から、閾値は2回に分割して転送しています。

    - ``val1`` は vth を右に6ビットシフト（=64で割る）して、head を足した値
    - ``val2`` は vth を左に2ビットシフト（=4をかける）して、下位8ビットを取り出した値

    詳細はこの関数のソースコードを確認してください。
    """
    val = vth
    head = 0b10000
    val1 = head + (val >> 6)
    val2 = (val << 2) & 0xFF

    logger.debug(f"Write: ch = {ch}")
    logger.debug(f"Write: val1 = {val1:b}")
    logger.debug(f"Write: val2 = {val2:b}")

    port.write(ch.to_bytes(1, "big"))
    port.write(val1.to_bytes(1, "big"))
    port.write(val2.to_bytes(1, "big"))

    read0 = port.readline().decode("utf-8", "ignore").strip()
    read1 = port.readline().decode("utf-8", "ignore").strip()
    read2 = port.readline().decode("utf-8", "ignore").strip()

    logger.debug(f"Read: ch = {read0}")
    logger.debug(f"Read: val1 = {read1}")
    logger.debug(f"Read: val2 = {read2}")

    success = False

    if read0 == "dame":
        success = False
        msg = f"Ch{ch}: Set threshold failed. Try again."
        logger.warning(msg)
    elif read0 == str(ch):
        success = True
        msg = f"Ch{ch}: Set threshold to {vth}."
        logger.success(msg)
    else:
        success = True
        msg = f"Ch{ch}: Set threshold to {vth}. Maybe."
        logger.success(msg)
    return success


def set_vth(daq: Daq, ch: int, vth: int) -> bool:
    """Set threshold

    チャンネル番号を指定し、閾値を設定します。
    チャンネル番号は1 - 3 の範囲で指定してください。
    閾値は1 - 1023 の範囲で指定してください。

    書き込みに成功すると ``success=True`` を返します。
    書き込みに失敗した場合は、警告メッセージを表示します。
    このとき、設定済みの閾値はそのままになります。

    :Args:
    - daq(Daq): DAQ設定オブジェクト
    - ch(int): チャンネル番号。1 - 3の範囲で指定してください
    - vth(int): 閾値の値。1 - 1023 の範囲で指定してください
    """

    # check values
    if ch not in range(1, 4):
        msg = f"value of ch is out of range: {ch}"
        logger.error(msg)
        sys.exit()

    if vth not in range(1, 1024):
        msg = f"value of vth is out of range: {vth}"
        logger.error(msg)
        sys.exit()

    mkdir_saved(daq)

    try:
        with open_serial_connection(daq) as port:
            success = write_vth(port, ch, vth)
            fname = Path(daq.saved) / daq.fname_logs

            with fname.open("a", newline="") as f:
                now = pendulum.now().to_iso8601_string()
                row = [str(now), str(ch), str(vth), str(success)]
                writer = csv.writer(f)
                writer.writerow(row)
            msg = f"Saved data to {fname}."
            logger.info(msg)
        return success
    except serial.SerialException as e:
        logger.error(e)
        msg = """Could not open the port. Device name might be wrong.
        Run 'arduino-cli board list' and check the device name.
        Edit 'daq.toml' if you needed.
        """
        logger.warning(msg)
        sys.exit()
    except Exception as e:  # noqa
        logger.error(e)
        msg = """
        Unaware error occurred.
        Please think if you need to handle this error.
        """
        sys.exit()


def set_vth_retry(daq: Daq, ch: int, vth: int, max_retry: int) -> bool:
    """
    Retry setting threshold.

    :Args:
    - daq (Daq): Daqオブジェクト
    - ch (int): チャンネル番号
    - vth (int): スレッショルド値
    - max_retry (int): 設定に失敗ししたときにトライする最大回数

    :Returns:
    - success (bool): スレッショルド値の設定の結果
    """

    for i in range(max_retry):
        success = set_vth(daq, ch, vth)
        if success:
            return True
        msg = f"Retry: {i} / {max_retry} times."
        logger.warning(msg)

    return False


def _read_event(port: serial.Serial) -> RealEvent:
    """OSECHIから1イベントを読み出す

    OSECHIが接続されているポートからデータを読み出します。
    引数に指定するポートは、あらかじめ開いたものを渡してください。
    ``run_daq``や``time_daq``でデータを取得するために使います。

    :Args:
    - port (serial.Serial): Serialオブジェクト

    :Returns:
    - event (RealEvent): 読み出した時刻を追加したデータ

    :Examples:
    ```python
    >>> with open_serial_connection() as port:
    >>>     event = _read_event(port)
    >>> event
    [pendulum.now(), top, mid, btm, adc, tmp, atm, hmd]
    ```

    """
    # logger.debug("[_read_event]")
    data = port.readline().decode("UTF-8").strip().split()
    if len(data) == 0:
        msg = f"No data to readout. Timed-out: {port.timeout}"
        logger.warning(msg)
    event = RealEvent()
    event.timestamp = pendulum.now()
    event.top = int(data[0])
    event.mid = int(data[1])
    event.btm = int(data[2])
    event.adc = int(data[3])
    event.tmp = float(data[4])
    event.atm = float(data[5])
    event.hmd = float(data[6])
    return event


def _loop_events_for_rows(port: serial.Serial, max_rows: int) -> Generator[RealEvent, None, None]:
    """イベント取得ループ（回数指定）

    指定した回数内にOSECHIからイベントを読み取るジェネレーター関数

    :Args:
    - `port (serial.Serial)`: 接続済みのSerialオブジェクト
    - `max_rows (int)`: イベントを読み取る最大回数

    :Yields:
    - `event (RealEvent)`: OSECHIの測定データ

    """
    logger.success("[_loop_events_for_rows]")
    rows = range(max_rows)
    for _ in tqdm(rows, leave=False, desc="loops"):
        event = _read_event(port)
        yield event


def _loop_events_for_duration(
    port: serial.Serial, max_duration: int
) -> Generator[RealEvent, None, None]:
    """イベント取得ループ（時間指定）

    指定した時間内にOSECHIからイベントを読み取るジェネレーター関数

    :Args:
    - `port (serial.Serial)`: 接続済みのSerialオブジェクト
    - `max_duration (int)`: イベントを読み取る最大時間（秒）

    :Yields:
    - `event (RealEvent)`: OSECHIの測定データ

    """
    logger.success("[_loop_events_for_duration]")
    start = pendulum.now()
    stop = start.add(seconds=max_duration)

    logger.debug(f"- DAQ Started: {start}")
    logger.debug(f"- DAQ Stop   : {stop}")

    while pendulum.now() < stop:
        event = _read_event(port)
        yield event

    end = pendulum.now()
    diff = end - start
    elapsed_time = diff.in_seconds()
    logger.debug(f"- DAQ Closed : {end} / Elapsed: {elapsed_time} sec.")


def loop_and_save(fname: Path, generator: Generator) -> list[RealEvent]:
    """イベント保存ループ

    ファイル名を指定して、イベントを保存します。
    保存形式はファイル名の拡張子で判定します。
    有効な拡張子は``[".csv", ".dat", ".tsv", ".json", ".jsonl"]``です。

    イベントの取得方法は``generator``で指定します。
    有効なジェネレーターは``[_loop_events_for_rows, _loop_events_for_duration]``です。

    :Args:
    - fname (Path): データを追記するファイル名
    - generator (Generator): イベントの取得方法 [`_loop_events_for_rows``, ``_loop_events_for_duration``]

    :Returns:
    - events (list[RealEvent]): OSECHIの測定データのリスト

    """
    suffix = fname.suffix
    logger.success("[loop_and_save]")
    events = []
    with fname.open("x") as f:
        for event in generator:
            events.append(event)
            if suffix in [".csv"]:
                row = event.to_csv_string()
                f.write(row + "\n")
            if suffix in [".dat", ".tsv"]:
                row = event.to_tsv_string()
                f.write(row + "\n")
            if suffix in [".json", ".jsonl"]:
                row = event.model_dump_json()
                f.write(row + "\n")
            f.flush()
    return events


def run_daq(port: serial.Serial, daq: Daq) -> None:
    """DAQを実行する（回数指定）

    OSECHIが接続されたUSBポートとシリアル通信をして、データ取得する。
    指定したファイル数と行数をでループ処理する。

    """
    mkdir_saved(daq)
    max_files = daq.max_files
    for nfile in tqdm(range(max_files), desc="files"):
        savef = get_savef_with_timestamp(daq, nfile)
        msg = f"Saving data to: {savef}."
        logger.info(msg)
        logger.info("Press Ctrl-c to stop.")
        loop_and_save(
            fname=savef,
            generator=_loop_events_for_rows(port=port, max_rows=daq.max_rows),
        )
        msg = f"Saved data to: {savef}."
        logger.success(msg)


def scan_daq(args: Daq, fname: str, duration: int) -> pd.DataFrame:
    """DAQを実行する（時間指定）

    :Args:
    - args (Daq): Daqオブジェクト
    - duration (int): 測定時間を秒で指定

    :Returns:
    - data (pd.DataFrame): 測定結果のデータフレーム
    """

    with open_serial_connection(args) as port:
        mkdir_saved(args)

        logger.debug(f"Port opened : {port.name}")
        events = loop_and_save(
            fname=Path(fname),
            generator=_loop_events_for_duration(port=port, max_duration=duration),
        )
        msg = f"Saved data to: {fname}."
        logger.success(msg)

    logger.debug(f"Port closed : {port.name}")
    # data = pd.DataFrame(rows)
    return events


def run(port: serial.Serial, args: Daq):
    """メインのDAQ

    run_daqのラッパー。例外処理などで囲んだもの。

    :versionadded: `0.6.0`.
    """

    try:
        run_daq(port, args)
    except serial.SerialException as e:
        logger.error(e)
        msg = """Could not open the port. Device name might be wrong.
        Run 'arduino-cli board list' and check the device name.
        Edit 'daq.toml' if you needed.
        """
        logger.warning(msg)
        sys.exit()
    except KeyboardInterrupt as e:
        logger.warning(e)
        msg = """Quit."""
        logger.info(msg)
        sys.exit()
    except Exception as e:  # noqa
        logger.error(e)
        msg = """Exit.
        Unaware error occurred.
        Please think if you need to handle this error.
        """
        logger.error(msg)
        sys.exit()


"""削除予定の関数"""


@deprecated(version="0.15.3", reason="Will be deprecated. Use scan_daq instead.")
def time_daq(args: Daq, duration: int) -> pd.DataFrame:
    """（削除予定）測定時間を指定してDAQを走らせます。

    :Args:
    - args (Daq): Daqオブジェクト
    - duration (int): 測定時間を秒で指定

    :Returns:
    - data (pd.DataFrame): 測定結果のデータフレーム
    """

    rows = []
    with open_serial_connection(args) as port:
        mkdir_saved(args)

        logger.debug("Port opened.")
        daq_start = pendulum.now()
        daq_stop = daq_start.add(seconds=duration)

        logger.debug(f"- DAQ Started: {daq_start}")
        logger.debug(f"- DAQ Stop   : {daq_stop}")

        while pendulum.now() < daq_stop:
            row = read_serial_data(port)
            event = RealEvent()
            event.timestamp = pendulum.now()
            event.top = int(row[1])
            event.mid = int(row[2])
            event.btm = int(row[3])
            event.adc = int(row[4])
            event.tmp = float(row[5])
            event.atm = float(row[6])
            event.hmd = float(row[7])
            rows.append(event.model_dump())
    daq_end = pendulum.now()
    diff = daq_end - daq_start
    elapsed_time = diff.in_seconds()
    logger.debug(f"- DAQ Closed : {daq_end} / Elapsed: {elapsed_time} sec.")
    data = pd.DataFrame(rows)
    return data


@deprecated(
    version="0.14.0",
    reason="Will be deprecated. Use threshold.scan_threshold_by_channel instead.",
)
def scan_ch_vth(daq: Daq, duration: int, ch: int, vth: int) -> list:
    """（削除予定）Run threshold scan.

    :Args:
    - daq (Daq): Daqオブジェクト
    - duration (int): 測定時間（秒）
    - ch (int): 測定するチャンネル番号
    - vth (int): スレッショルド値

    :Returns:
    - data (list): [測定時刻、チャンネル番号、スレッショルド値、イベント数]のリストを返します。
    """

    logger.warning("Will be deprecated. Please use threshold.scan_by_channel instead.")

    # Try to set the threshold
    if not set_vth_retry(daq, ch, vth, 3):
        msg = f"Failed to set threshold: ch{ch} - {vth}"
        logger.error(msg)
        return []

    # Collect data
    try:
        rows = time_daq(daq, duration)
        counts = len(rows)
        tmp = rows["tmp"].mean()
        atm = rows["atm"].mean()
        hmd = rows["hmd"].mean()
        fname = get_savef_with_timestamp(daq, ch)
        rows.to_csv(fname, index=False)
        msg = f"Saved data to: {fname}"
        logger.info(msg)
    except Exception as e:
        msg = f"Failed to collect data due to: {str(e)}"
        logger.error(msg)
        counts = 0
        tmp = 0
        atm = 0
        hmd = 0

    # Save Summary
    now = pendulum.now().to_iso8601_string()
    data = [now, duration, ch, vth, counts, tmp, atm, hmd]
    fname = Path(daq.saved) / daq.fname_scan
    with fname.open("a", newline="") as f:
        writer = csv.writer(f)
        writer.writerow(data)
    msg = f"Added data to: {fname}"
    logger.info(msg)

    return data


@deprecated(
    version="0.14.0",
    reason="Will be deprecated. Use threshold.scan_thresholds instead.",
)
def scan_ch_thresholds(daq: Daq, duration: int, ch: int, thresholds: list[int]) -> list[list]:
    """（削除予定）Run threshold scan.

    :Args:
    - daq (Daq): Daqオブジェクト
    - duration (int): 測定時間（秒）
    - ch (int): チャンネル番号
    - thresholds (list[int]): スレッショルド値のリスト

    :Returns:
    - rows (list[list]): [測定時刻、チャンネル番号、スレッショルド値、イベント数]のリスト
    """

    logger.warning("Will be deprecated. Please use threshold.scan_thresholds instead.")
    # Estimated time for scan
    msg = f"Number of points: {len(thresholds)}"
    logger.info(msg)
    estimated_time = len(thresholds) * duration
    msg = f"Estimated time: {estimated_time} sec."
    logger.info(msg)

    # すべてのチャンネルの閾値を高くする
    set_vth_retry(daq, 1, 500, 5)
    set_vth_retry(daq, 2, 500, 5)
    set_vth_retry(daq, 3, 500, 5)

    rows = []
    n = len(thresholds)
    for i, vth in enumerate(thresholds):
        msg = "-" * 40 + f"[{i+1:2d}/{n:2d}: {vth}]"
        logger.info(msg)
        row = scan_ch_vth(daq, duration, ch, vth)
        if row:
            rows.append(row)

    return rows


@deprecated(
    version="0.15.0",
    reason="Will be deprecated. Use _read_event (RealEvent).",
)
def _read_serial_data_as_list(port: serial.Serial) -> list:
    """（削除予定）Read serial data from port.

    OSECHIが接続されているポートからデータを読み出します。
    引数に指定するポートは、あらかじめ開いたものを渡してください。
    ``run_daq``や``time_daq``でデータを取得するために使います。

    :Args:
    - port (serial.Serial): Serialオブジェクト

    :Returns:
    - row (list): 読み出した時刻を追加したデータのリスト

    :Examples:
    ```python
    >>> with open_serial_connection() as port:
    >>>     row = read_serial_data(port)
    >>>     row
    [日付, top, mid, btm, adc, tmp, atm, hmd]
    ```

    """
    msg = "Will be deprecated. Use _read_event."
    logger.warning(msg)
    now = pendulum.now().to_iso8601_string()
    data = port.readline().decode("UTF-8").strip()
    if len(data) == 0:
        msg = f"No data to readout. Timed-out: {port.timeout}"
        logger.warning(msg)
    row = f"{now} {data}".split()
    return row


@deprecated(
    version="0.15.0",
    reason="Will be deprecated. Use _read_event (RealEvent).",
)
def read_serial_data(port: serial.Serial) -> list:
    """（削除予定）"""
    msg = "Will be deprecated. Use _read_event."
    logger.warning(msg)
    data = _read_serial_data_as_list(port)
    return data


@deprecated(version="0.15.3", reason="Will be deprecated. Use loop_and_save instead.")
def _loop_and_save_events(
    fname: Path, port: serial.Serial, max_rows: int, suffix: str = ".csv"
) -> list:
    """（削除予定）イベント保存ループ

    :Args:
    - `f (typing.TextIO)`: 開いたファイルオブジェクト
    - `port (serial.Serial)` : 接続済みのSerialオブジェクト
    - `max_rows (int)`: 1ファイルあたりの行数の最大値
    - `suffix (str)`: ファイルの拡張子

    :Return:
    - `events (list)`: 複数イベントの測定データ

    """
    logger.warning("Will be deprecated. Use loop_and_save instead.")
    events = []
    with fname.open("x") as f:
        for event in _loop_events_for_rows(port, max_rows):
            events.append(event)
            if suffix in [".csv"]:
                row = event.to_csv_string()
                f.write(row + "\n")
            if suffix in [".dat", ".txv"]:
                row = event.to_tsv_string()
                f.write(row + "\n")
            if suffix in [".json", ".jsonl"]:
                row = event.model_dump_json()
                f.write(row + "\n")
            f.flush()
    return events


@deprecated(version="0.15.3", reason="Will be deprecated. Use _loop_and_save_events instead.")
def loop_and_save_events(f: TextIO, daq: Daq, port: serial.Serial) -> list[str]:
    """（削除予定）イベント保存ループ

    :Args:
    - f (TextIO): データを書き込むファイルオブジェクト
    - daq (Daq): Daqオブジェクト
    - port (serial.Serial): 接続済みのSerialオブジェクト

    """
    logger.warning("Will be deprecated. Use _loop_and_save_events instead.")
    rows = daq.max_rows
    events = []
    for event in _loop_events_for_rows(port, max_rows=rows):
        events.append(event.model_dump_json())
        if daq.suffix in [".csv"]:
            row = event.to_csv_string()
            f.write(row + "\n")
        elif daq.suffix in [".dat", ".tsv"]:
            row = event.to_tsv_string()
            f.write(row + "\n")
        elif daq.suffix in [".json", ".jsonl"]:
            row = event.model_dump_json()
            f.write(row + "\n")
        f.flush()
    return events


@deprecated(version="0.15.0", reason="Will be deprecated. Use _loop_and_save_events instead.")
def save_serial_data(f, daq: Daq, port: serial.Serial) -> list:
    """（削除予定）

    :Args:
    - f: ファイルポインタ
    - daq (Daq): Daqオブジェクト
    - port (serial.Serial): Serialオブジェクト

    :Return:
    - rows (list[list]): 取得したデータのリスト

    :TODO:
    - Daqオブジェクトに依存しない関数にしたい（ジェネレーターにするのがいいのかな？）
    - pd.DataFrameを返した方がいいかもしれない？
    """
    logger.warning("Will be deprecated. Use _loop_and_save_events instead.")
    max_rows = daq.max_rows
    rows = []
    for _ in tqdm(range(max_rows), leave=False, desc="rows"):
        row = read_serial_data(port)
        rows.append(row)
        if daq.suffix == ".csv":
            writer = csv.writer(f)
            writer.writerow(row)
        else:
            writer = csv.writer(f, delimiter=" ")
            writer.writerow(row)
        f.flush()
    return rows



import sys
from pathlib import Path
from typing import Any

import pandas as pd
import pendulum
from loguru import logger
from pydantic import BaseModel
from deprecated import deprecated


class PathSettings(BaseModel):
    """データセットのパス設定

    :TODO:
    - データセットの配置を変更したら、パスのデフォルト値を変更する

    """

    drive: str = "."
    """データセットのルートディレクトリ（デフォルト="."）"""

    def raw_data_path(self, path: str = "raw_data") -> Path:
        """測定データのパス"""
        p = Path(self.drive) / path
        if not p.exists():
            msg = f"Path does not exist. {p}"
            logger.error(msg)
            p = Path(self.drive)
        return p

    def preprocessed_data_path(self, path: str = "parsed") -> Path:
        """前処理したデータのパス"""
        p = Path(self.drive) / path
        if not p.exists():
            msg = f"Path does not exist. {p}"
            logger.error(msg)
            p = Path(self.drive)
        return p

    def resampled_data_path(self, path: str = "parsed") -> Path:
        """リサンプルしたデータのパス"""
        p = Path(self.drive) / path
        if not p.exists():
            msg = f"Path does not exist. {p}"
            logger.error(msg)
            p = Path(self.drive)
        return p


class Run(PathSettings):
    """ラン設定

    :Examples:

    ```python
    run = Run(run_id=65, date="20240521")
    data1 = run.preprocessed_data()
    data2 = run.resample_data()
    data3 = run.ondotori_data()
    ```

    実行する場所からデータセットまでのパスを``drive``で変更できます。

    ```python
    run = Run(run_id=65, date="20240521", drive="../data")
    data = run.resample_data()
    ```

    """

    run_id: int
    """ラン番号"""
    date: str
    """測定日"""
    name: str | None = None
    """ファイル名"""

    def model_post_init(self, __context: Any) -> None:
        if self.name is None:
            self.name = self._stem()
        return super().model_post_init(__context)

    def _stem(self) -> str:
        """（プライベート関数）ファイル名のstemを取得する"""
        stem = f"{self.date}_run{self.run_id}"
        return stem

    def _osechi_raw_data(self, search_pattern="osechi_data_*.csv") -> list[Path]:
        """（プライベート関数）前処理したファイル名のリストを取得する"""
        p = self.raw_data_path() / self.name
        fnames = sorted(p.glob(search_pattern))
        return fnames

    def _osechi_preprocessed_data(self) -> list[Path]:
        """（プライベート関数）前処理したファイル名のリストを取得する"""
        p = self.preprocessed_data_path()
        if not p.exists():
            msg = f"Path does not exist. {p}"
            logger.error(msg)
        fnames = sorted(p.glob(f"{self._stem()}.csv.gz"))
        return fnames

    def _osechi_resampled_data(self) -> list[Path]:
        """（プライベート関数）リサンプルしたファイル名のリストを取得する"""
        p = self.resampled_data_path()
        fnames = sorted(p.glob(f"{self._stem()}.csv"))
        return fnames

    def _ondotori_raw_data(self) -> list[Path]:
        """（プライベート関数）おんどとりのデータを取得する"""
        p = self.raw_data_path() / self._stem()
        fnames = sorted(p.glob("TR41_*.csv"))
        return fnames

    def _threshlod_logs_data(self) -> list[Path]:
        """（プライベート関数）スレッショルド設定のログを取得する"""
        p = self.raw_data_path() / self._stem()
        fnames = sorted(p.glob("threshold_logs.csv"))
        return fnames

    def _to_dataframe(self, fnames, **kwargs) -> pd.DataFrame:
        """（プライベート関数）すべてのファイルをデータフレームに変換する

        :Args:
        - `fnames (list[Path])`: ファイル名のリスト
        - `**kwargs`: pd.DataFrameのオプション

        :Returns:
        - `data (pd.DataFrame)`: データフレーム
        """
        dfs = []
        for fname in fnames:
            df = pd.read_csv(fname, **kwargs)
            if not df.empty:
                dfs.append(df)
        if len(dfs) > 0:
            data = pd.concat(dfs, ignore_index=True)
        else:
            data = pd.DataFrame()
        return data

    def preprocessed_data(self) -> pd.DataFrame:
        """前処理したデータフレーム"""
        fnames = self._osechi_preprocessed_data()
        data = self._to_dataframe(fnames, parse_dates=["time"])
        return data

    def resampled_data(self) -> pd.DataFrame:
        """リサンプルしたデータフレーム"""
        fnames = self._osechi_resampled_data()
        data = self._to_dataframe(fnames, parse_dates=["time"])
        return data

    def ondotori_data(self) -> pd.DataFrame:
        """おんどとりのデータフレーム"""
        fnames = self._ondotori_raw_data()
        data = self._to_dataframe(
            fnames, skiprows=4, names=["datetime", "tmpC"], parse_dates=["datetime"]
        )
        return data

    def threshold_logs(self) -> pd.DataFrame:
        """スレッショルド設定のデータフレーム"""
        fnames = self._threshlod_logs_data()
        data = self._to_dataframe(
            fnames, names=["time", "ch", "vth", "success"], parse_dates=["time"]
        )
        return data

    def __str__(self):
        return self.name


def reformat_datetime(datetime: str) -> str:
    """
    日時を変換する

    :Examples:

    ```python
    >>> reformat_datetime("2022-05-18T20:48:14")
    2022-05-18T20:48:14

    >>> reformat_datetime("2022-06-02T18:52:36.442190+09:00")
    2022-06-02T18:52:36
    ```

    """
    dt = pendulum.parse(datetime)
    fmt = dt.format("YYYY-MM-DDTHH:mm:ss")
    return fmt


@deprecated(
    version="0.15.0",
    reason="削除する予定の関数です。Runオブジェクトを使ってみてください。",
)
def load_raw_data(fname: Path, **kwargs) -> pd.DataFrame:
    """
    DAQで取得したデータ形式を ``pd.DataFrame`` に変換して読み込む

    :Args:
    - ``fname (Path)``: 測定データのファイル名
    - ``**kwargs``: ``pd.read_csv``のオプション


    :Returns:
    - ``data (pd.DataFrame)``: データフレーム

    :Notes:
    - ファイルの保存形式（＝拡張子）で ``pd.read_csv`` の引数をちょっと変える必要がある。拡張子は ``.dat`` / ``.csv`` だけOKにしてあり、それ以外の場合は ``sys.exit`` する。
    - ``.csv`` はそのままカンマ区切り、``.dat`` はスペース区切り（ ``sep=" "`` ）として、 ``pd.read_csv`` する。
    - ファイル内のカラム名は適当だったり、なかったりする。適切なカラム名を付与する。カラム名はこの関数内にハードコードしている。
    - イベント時刻（ ``time`` ）は ``pd.datetime`` オブジェクトに変換する。使ったDAQのバージョンによって記録された日時の形式が異なるので、それを内部で変換している。
    - 各レイヤーのヒットの有無（ ``True`` / ``False`` ）を確認して、ヒット用のカラム（ ``hit_top`` / ``hit_mid`` / ``hit_btm`` ）に保存する。現在、各レイヤーの値自体には意味がない。そのうち光量など意味を持たせる可能性はあるかも？
    - ヒットのあったレイヤーのパターンを計算して 8ビット で表現する。``hit_type = hit_top * 4 + hit_mid * 2 + hit_btm * 1``
    """

    _names = ["time", "top", "mid", "btm", "adc", "tmp", "atm", "hmd"]
    suffix = fname.suffix
    if suffix == ".dat":
        data = pd.read_csv(fname, names=_names, sep=" ", comment="t", **kwargs)
    elif suffix == ".csv":
        data = pd.read_csv(fname, names=_names, **kwargs)
    else:
        error = f"Unknown suffix : {suffix}"
        logger.error(error)
        sys.exit()

    data["time"] = data["time"].apply(reformat_datetime)
    data["time"] = pd.to_datetime(data["time"], format="ISO8601")
    # 各レイヤーのヒットの有無を確認する
    data["hit_top"] = data["top"] > 0
    data["hit_mid"] = data["mid"] > 0
    data["hit_btm"] = data["btm"] > 0
    # ヒットのあったレイヤーのパターンを8ビットで表現する
    data["hit_type"] = data["hit_top"] * 4 + data["hit_mid"] * 2 + data["hit_btm"] * 1
    return data


@deprecated(
    version="0.15.0",
    reason="削除する予定の関数です。Runオブジェクトを使ってみてください。",
)
def load_files(fnames: list[Path]) -> pd.DataFrame:
    """複数のファイルを読み込み、単一のデータフレームに変換する

    :Args:
    - ``fnames (list[Path])``: 測定データのファイルの一覧

    :Returns:
    - ``data (pd.DataFrame)``: すべてのファイルを結合したデータフレーム

    :Notes:
    - ``fnames`` で列挙されたファイルごとに、データフレームに変換する
    - ファイル名の一覧が空の場合は終了する
    - 個々のデータフレームを結合して単一のデータフレームを作成する
    - 結合したデータフレームは、時刻（``time``）でソートしておく
    """
    if len(fnames) == 0:
        error = "No files listed"
        logger.error(error)
        sys.exit()

    _loaded = []
    for fname in fnames:
        _data = load_raw_data(fname)
        _loaded.append(_data)
    loaded = pd.concat(_loaded, ignore_index=True)
    loaded = loaded.sort_values(["time"])
    debug = f"Entries : {len(loaded)}"
    logger.debug(debug)
    return loaded
